local MAJOR = "LibScriptableDisplayBuffer-1.0"
local MINOR = 14
assert(LibStub, MAJOR.." requires LibStub")
local LibBuffer = LibStub:NewLibrary(MAJOR, MINOR)
if not LibBuffer then return end
local LibError = LibStub("LibScriptableDisplayError-1.0")
assert(LibError, MAJOR .. " requires LibScriptableDisplayError-1.0")

local pool = setmetatable({}, {__mode = "k"})

if not LibBuffer.__index then
	LibBuffer.__index = LibBuffer
end

local new, del
do
	local pool = setmetatable({}, {__mode = "k"})
	function new(size)
		local obj = next(pool)
		if obj then
			local smallest = obj
			for tbl in pairs(pool) do
				if #tbl >= size and #tbl < #smallest then
					smallest = tbl
				end
			end
			obj = smallest
			pool[obj] = nil
		else
			obj = {}
		end
		return obj
	end
	function del(obj)
		pool[obj] = true
	end
end

-- @name LibScriptableDisplayBuffer:New
-- @usage LibScriptableDisplayBuffer:New(name, size, defval, errorLevel)
-- @param name A name for the buffer
-- @param size The buffer's size
-- @param defval The buffer's default value. The alpha channel.
-- @param errorLevel The verbocity level
-- @return A new LibScriptableDisplayBuffer object

function LibBuffer:New(name, size, defval, errorLevel)
	assert(type(name) ~= "nil", "LibBuffer requires a name")
	assert(type(size) ~= "nil", "Buffer size is required")
	local obj = next(pool)

	if obj then
		pool[obj] = nil
	else
		obj = {}
	end

	obj.buffer = new(size)

	setmetatable(obj, self)

	obj.name = name
	obj.defval = defval or " "
	obj.error = LibError:New(MAJOR .. " " .. name, errorLevel)
	obj:Resize(size)

	return obj

end

-- @name LibScriptableDisplayBuffer:Del
-- @usage LibScriptableDisplayBuffer:Del([lcd]) or object:Del()
-- @param lcd An optional lcd object
-- @return Nothing
function LibBuffer:Del(buff)
	if not buff then buff = self end
	pool[buff] = true
	buff.error:Del()
	for i = 0, buff:Size() do
		buff.buffer[i] = false
	end
	buff.size = 0
	del(buff.buffer)
end

-- @name :Wipe
-- @usage :Wipe()
-- @return Nothing
function LibBuffer:Wipe()
	for i = 0, self:Size() - 1 do
		self.buffer[i] = self.defval
	end
end

-- @name :Fill
-- @usage :Fill(val)
-- @param val The value to fill the buffer with
-- @return Nothing
function LibBuffer:Fill(val)
	for i = 0, self:Size() - 1 do
		self.buffer[i] = val
	end
end

-- @name :Resize
-- @usage :Resize(size)
-- @param size The buffer's new size
-- @return Nothing
function LibBuffer:Resize(size)
	assert(self.defval ~= nil)
	assert(type(size) == "number")
	if self:Size() < size then
		for i = self:Size(), size - 1 do
			self.buffer[i] = self.defval
		end
	end

	if self:Size() > size then
		for i = size - 1, self:Size() do
			self.buffer[i] = false
		end
	end

	self.size = size

	if size > 0 then
		for i = 0, size - 1 do
			assert(self.buffer[i] ~= nil, format("resize %d %d", i, size))
		end
	end
end

function LibBuffer:Replace(pos, val)
	assert(type(pos) == "number", "Position sent to Replace isn't a number")
	assert(val ~= nil, "Val sent to Replace is nil")
	if pos > self.size then self.size = pos end
	self.buffer[pos] = val
end

function LibBuffer:Clear()
	for i = 0, self:Size() - 1 do
		self.buffer[i] = self.defval
	end
end

local pool = setmetatable({}, {__mode="k"})
local function new()
	local obj = next(pool)

	if obj then
		pool[obj] = nil
	else
		obj = {}
	end

	wipe(obj)
	return obj
end

local function del(obj)
	pool[obj] = true
end

function LibBuffer:AsString()
	assert(type(self.defval) == "string")
	local tbl = new()
	for i = 0, self:Size() - 1 do
		if type(self.buffer[i]) == "string" then
			tbl[i + 1] = self.buffer[i]
		else
			tbl[i + 1] = ""
		end
	end
	local str = table.concat(tbl)
	del(tbl)

	--local str = (self.buffer[0] or "") .. table.concat(self.buffer)
	--[[
	for i = 0, self:Size() - 1 do
		if self.buffer[i] == nil then break end
		if type(self.buffer[i]) == "string" then
			str = str .. self.buffer[i]
		else
			str = str .. self.defval
			self.error:Print(":AsString() - Bad index: " .. i .. ", type: " .. type(self.buffer[i]) .. ", size: " .. self:Size())
		end
	end]]
	return str
end

function LibBuffer:FromString(str)
	if type(str) ~= "string" then return end
	for i = 0, strlen(str) - 1 do
		self.buffer[i] = str:sub(i + 1, i + 1)
	end
	if strlen(str) < self:Size() then
		for i = strlen(str) - 1, self:Size() do
			self.buffer[i] = false
		end
	end
	self.size = strlen(str)
end

function LibBuffer:MergeString(pos1, str, pos2, length)
	for n = 0, length - 1 do
		self.buffer[pos1 + n] = str:sub(pos2, pos2 + length)
	end
end

function LibBuffer:Size()
	--[[
	local count = 0
	for k in pairs(self.buffer) do
		count = count + 1
	end
	return count
	]]
	return self.size or 0
end

function LibBuffer:Sub(num1, num2)

	if not num2 then num2 = self:Size() - 1 end
	if num1 > num2 or num2 > self:Size() - 1 then return end
	
	local t = LibBuffer:New(self.name .. ":" , num2 - num1 + 1, self.defval, self.errorLevel)
	for i = num1, num2 do
		t.buffer[i - num1] = self.buffer[i] or self.defval
	end
	return t
end

function LibBuffer:Copy(other)
	table.wipe(self.buffer)
	for k, v in pairs(other.buffer) do
		self.buffer[k] = other.buffer[k]
	end
end

function LibBuffer:Merge(from, pos, len)
	local i = 0
	while i < len - (self.size - len) do
		assert(from.buffer[i] ~= nil)
		self.buffer[i + pos] = from.buffer[i]
		i = i + 1
	end
end

function LibBuffer:Memset(pos, char, length)
	for n = pos, pos + length - 1 do
		self.buffer[n] = char
	end
end

function LibBuffer:Memcopy(pos1, buffer, pos2, length)
	for n = 0, length - 1 do
		self.buffer[pos1 + n] = buffer.buffer[pos2 + n] or self.defval
	end
end

function LibBuffer:MovingAverageExp(alpha, epsilon)
	if type(epsilon) ~= "number" then epsilon = 0 end

	if not (0 < alpha and alpha < 1) then
		return
	end

	if not (0 <= epsilon and epsilon < alpha) then
		return
	end
	
	result = LibBuffer:New("result", self:Size(), 0, self.errorLevel)

	for i = 0, result:Size() - 1 do
		local currentWeight = 1.0
		local numerator = 0
		local denominator = 0
		local sub = self:Sub(i)
		for j = 0, sub:Size() - 1 do
			local value = sub.buffer[j]
			numerator = numerator + value * currentWeight
			denominator = denominator + currentWeight

			currentWeight = currentWeight * alpha
			if currentWeight < epsilon then break end
		end
		result.buffer[i] = numerator / denominator
		sub:Del()
	end

	return result
end
